-- BecaEnergy Wi-Fi Thermostat Plug-In for MicasaVerde Vera

-- (c) Martyn Barnett 2017
-- (heavily leveraged and inherited from Alan Carters Heatmiser Plugin (thank you Alan))



local	BECAENERGY_VERSION 	= "1.90"



local	TEMPERATURE_SERVICE = "urn:upnp-org:serviceId:TemperatureSensor1"

local	TEMPERATURE_SETPOINT= "urn:upnp-org:serviceId:TemperatureSetpoint1"

local	HOME_AUTO_SERVICE	= "urn:micasaverde-com:serviceId:HaDevice1"

local 	BECAENERGY_SERVICE 	= "urn:m-barnett-com:serviceId:BecaEnergy"

local 	BECAENERGY_LOG 		= "BecaEnergy: "

local	HVAC_USER_SERVICE	= "urn:upnp-org:serviceId:HVAC_UserOperatingMode1"

local	SECURITY_SERVICE	= "urn:micasaverde-com:serviceId:SecuritySensor1"



local		FUNCTION_READ		= 160

local		FUNCTION_WRITE		= 161

local		FUNCTION_TIME		= 168

local 	OR, XOR, AND = 1, 3, 4

local		DCBPNT_START		= 0

local		DCBPNT_ALL			= 65535



local		DCBPNT_MODEL		= 4

local		DCBPNT_TEMP_FORMAT	= 5

local		DCBPNT_SENSOR_SEL		= 13

local		DCBPNT_FROSTTEMP		= 17

local		DCBPNT_SETTEMP		= 166

local		DCBPNT_ONOFF		= 162

local		DCBPNT_KEYLOCK		= 22

local		DCBPNT_RUNMODE		= 169

local		DCBPNT_TEMPHOLDMINS	= 31

local		DCBUNQ_TEMPHOLDMINS 	= 32

local		DCBPNT_REMOTETEMP		= 33

local		DCBPNT_FLOORTEMP		= 35

local		DCBPNT_ROOMTEMP		= 37

local		DCBPNT_ERRORCODE		= 39	-- Not currently used

local		DCBPNT_HEATSTATE		= 169

local		DCBPNT_HWSTATE		= 43	-- HW models only

local		DCBUNQ_HWSTATE		= 42	-- HW models only

local		DCBUNQ_SETTIME		= 168



local 	ipAddress

local 	ipPort 				= 8899

local 	errorsCRC

local		comm_failure



local 	message_type		= ""

local		becaenergy_poll		= "-"

local		becaenergy_cmd		= "*"

local 	incomingDCB 		= ""



local	socket			= require("socket")

local	bit 				= require("bit")

local	err

local 	lug_device 			= nil

local	model				= 0

local	model_text			= ""

local ModeStatus 			= "AutoChangeOver"

local	tempUnits			= ""

local	lo_temp_limit = 5

local	hi_temp_limit = 35

local	lo_frost_limit = 7

local	hi_frost_limit = 17

local	timeSyncEnable

local	loggingEnable

local	setHoldMins

local	setHoldHours

local	setHoldTemp

local	pollFString			=	""

local	pollFrequency		=	30

local	timeSHString		=	""

local	timeSyncHour		=	3



local	last_tempFormat		=	0

local	last_tempHold 		=	0

local	last_onoff_status	=	0

local	last_keylock_status	=	0

local	last_runmode_status	=	0

local	last_sensor_sel		=	0

local	last_frost_temp		=	0

local	last_heat_state		=	0

local	last_hw_state		=	0

local	last_cur_temp		=	0

local	last_set_temp		=	0

local	last_mode_status	=	""

local	hs_change			=	0

local	cs_change			=	1

local	sensor				=	"Unsupported Mode"

local start_of_incoming_data = 0

local FullDCB			= ""


	--
	-- called by InComing for a received character	
	--
	function BecaIncoming(incomingDCB)

		if (string.byte(incomingDCB) == 0x50) then
			start_of_incoming_data = 0
			FullDCB = incomingDCB
		else
			start_of_incoming_data = start_of_incoming_data + 1
			FullDCB = FullDCB .. incomingDCB
		end

		-- Have we got a whole packet yet?
		if (start_of_incoming_data == 7) then
		
			-- yes - we have a whole packet

			-- check that packet for validity
			-- remember the format
			-- 0x50 REPLY
			-- 0x01 ID1
			-- 0x01 ID2
			-- <data 0> <BIT4 - ON/OFF	 BIT7 - Heat ON/OFF
			-- <data 1> TEMPERATURE CALIBRATION
			-- <data 2> TEMPERATURE SETTING
			-- <data 3> AMBIENT TEMPERATURE 
			-- <CRC>
			
			if (string.byte(string.sub(FullDCB, 1, 1)) == 0x50 and
				string.byte(string.sub(FullDCB, 2, 2)) == 0x01 and
				string.byte(string.sub(FullDCB, 3, 3)) == 0x01 ) then

				-- YES! it is a valid packet

				-- now go check the CRC just to be sure its a good packet
				if (string.byte(string.sub(FullDCB, 8, 8)) == CalculateCRC(FullDCB)) then

					becaenergyLogDCB(DCBtoString(FullDCB), "<--", message_type)
					processDCB(FullDCB)
				end
			end
		end
	end

	--
	-- Run once at Luup engine startup.
	--
	function initialise(lul_device)

		luup.log(BECAENERGY_LOG.."Initialising BecaEnergy Device")

		lug_device = lul_device

		local ip_missing = 0

			if (luup.devices[lul_device].ip == "") then

			luup.task(BECAENERGY_LOG.."Assign IP Address", 2, string.format("%s[%d]", luup.devices[lul_device].description, lul_device), -1)

			luup.log(BECAENERGY_LOG.."No IP Address")

			ip_missing = 1

			end

			if (ip_missing == 1) then

				return false

			end

			ipAddress = luup.devices[lul_device].ip

			luup.log(BECAENERGY_LOG.."Wi-Fi Device on "..ipAddress..":"..ipPort)

			luup.variable_set(BECAENERGY_SERVICE, "Version", BECAENERGY_VERSION, lul_device)

			luup.variable_set(HOME_AUTO_SERVICE, "CommFailure", "0", lul_device)

			comm_failure = 0
			
			luup.variable_set(BECAENERGY_SERVICE, "CRCErrors", "0", lul_device)

			errorsCRC = 0

			-- Open the TCP Port
			luup.io.open(lul_device, ipAddress, ipPort)
			if( luup.io.is_connected(lul_device) == false ) then
				luup.log(BECAENERGY_LOG.."Cannot Connect BecaEnergy at "..ipAddress..":"..ipPort)
			else
				luup.log(BECAENERGY_LOG.."Connected to BecaEnergy at "..ipAddress..":"..ipPort)
			end
			
			if (luup.variable_get(BECAENERGY_SERVICE, "PollFrequency", lul_device) == nil) then

			luup.variable_set(BECAENERGY_SERVICE, "PollFrequency", "30", lul_device)

			pollFrequency = 30

			else

			pollFString = luup.variable_get(BECAENERGY_SERVICE, "PollFrequency", lul_device)

			pollFrequency = tonumber(pollFString)

			if (pollFrequency > 0) and (pollFrequency < 30) then

				pollFrequency = 30

				luup.log(BECAENERGY_LOG.."Invalid Poll Frequency - will use default")

			end

			end

						

			if (luup.variable_get(BECAERGY_SERVICE, "SetHoldMins", lul_device) == nil) then

			luup.variable_set(BECAENERGY_SERVICE, "SetHoldMins", "0", lul_device)

			setHoldMins = 0

		else

			setHoldMins = luup.variable_get(BECAENERGY_SERVICE, "SetHoldMins", lul_device)

			end

			

			if (luup.variable_get(BECAENERGY_SERVICE, "SetHoldHours", lul_device) == nil) then

			luup.variable_set(BECAENERGY_SERVICE, "SetHoldHours", "1", lul_device)

			setHoldHours = 1

		else

			setHoldHours = luup.variable_get(BECAENERGY_SERVICE, "SetHoldHours", lul_device)

			end

			

		if (luup.variable_get(BECAENERGY_SERVICE, "SetHoldTemp", lul_device) == nil) then

			luup.variable_set(BECAENERGY_SERVICE, "SetHoldTemp", "20", lul_device)

			setHoldTemp = 20

		else

			setHoldTemp = luup.variable_get(BECAENERGY_SERVICE, "SetHoldTemp", lul_device)

			end

			

			if (luup.variable_get(BECAENERGY_SERVICE, "TimeSyncEnable", lul_device) == nil) then

			luup.variable_set(BECAENERGY_SERVICE, "TimeSyncEnable", "0", lul_device)

			timeSyncEnable = "0"

		else

			timeSyncEnable = luup.variable_get(BECAENERGY_SERVICE, "TimeSyncEnable", lul_device)

			end



			if (luup.variable_get(BECAENERGY_SERVICE, "TimeSyncHour", lul_device) == nil) then

			luup.variable_set(BECAENERGY_SERVICE, "TimeSyncHour", "3", lul_device)

			timeSyncHour = 3

		else

			timeSHString = luup.variable_get(BECAENERGY_SERVICE, "TimeSyncHour", lul_device)

			timeSyncHour = tonumber(timeSHString)

			if (timeSyncHour < 0) or (timeSyncHour > 23) then

				timeSyncHour = 3

				luup.variable_set(BECAENERGY_SERVICE, "TimeSyncHour", "3", lul_device)

				luup.log(BECAENERGY_LOG.."Invalid Time Sync Hour - will use default")

			end

			end

						

			if (luup.variable_get(BECAENERGY_SERVICE, "LoggingEnable", lul_device) == nil) then

			luup.variable_set(BECAENERGY_SERVICE, "LoggingEnable", "1", lul_device)

			loggingEnable = 1

		else

			loggingEnable = luup.variable_get(BECAENERGY_SERVICE, "LoggingEnable", lul_device)

			end

			

			if (luup.variable_get(BECAENERGY_SERVICE, "TempUnits", lul_device) == nil) then

			luup.variable_set(BECAENERGY_SERVICE, "TempUnits", "", lul_device)

			tempUnits = ""

		else

			tempUnits = luup.variable_get(BECAENERGY_SERVICE, "TempUnits", lul_device)

			end

		

		if pollFrequency > 0 then

			luup.variable_set(BECAENERGY_SERVICE, "HeatStateText", "", lug_device)

			luup.call_timer('routinePoll', 1, "2", "")

		else

			luup.variable_set(BECAENERGY_SERVICE, "HeatStateText", "DISABLED", lug_device)

		end

		

		if (pollFrequency > 0 and timeSyncEnable == "1") then

			luup.call_timer('setBecaEnergyTime', 1, "10", "")

		end

		cs_change = 1

		luup.set_failure(false)

			luup.log(BECAENERGY_LOG.."Startup complete")

	end


	-- Poll device at specified poll frequency

	function routinePoll()

		luup.call_timer('routinePoll', 1, pollFrequency, "")

		message_type = becaenergy_poll

		frameSend(frameGenerate(FUNCTION_READ, 0, 0, 0, 0))

	end


	-- Sync BecaEnergy time with Vera

	function setBecaEnergyTime()

		local Data1 = 0
		local Data2 = 0
		local Data3 = 0
		local Data4 = 0

		luup.call_timer('setBecaEnergyTime', 1, "1h", "")

		local Now = os.date("*t")

		if Now.hour == timeSyncHour then

			Now.year = Now.year - 2000

			Now.wday = Now.wday - 1

			if Now.wday == 0 then

				Now.wday = 7

			end

			Data1 = (Now.sec)
			Data2 = (Now.min)
			Data3 = (Now.hour)
			Data4 = (Now.wday)

			message_type = becaenergy_cmd

			frameSend(frameGenerate(FUNCTION_TIME, Data1, Data2, Data3, Data4))

			luup.log(BECAENERGY_LOG.."Time sync complete")

		end

	end


	-- BecaEnergy On/Off

	function setOnOff(lul_device,lul_settings)
		
		local Data1 = 0
		local Data2 = 0xFC 
		local Data3 = (luup.variable_get(BECAENERGY_SERVICE, "SetHoldTemp", lul_device) * 2)
		local Data4 = 0

		local State = tonumber(lul_settings['State'])

		if(State ~= 0) then

			State = 1

	 	end

		if State == 1 then Data1 = 0x18
		else Data = 0x08
		end

		message_type = becaenergy_cmd

		frameSend(frameGenerate(FUNCTION_WRITE, Data1, Data2, Data3, Data4))

		luup.log(BECAENERGY_LOG.."Device " .. lul_device .. ": SetOnOff " .. lul_settings['State'])

	end

	
	-- Send new temperature setpoint to BecaEnergy

	function setCurrentSetpoint(lul_device,lul_settings)

		local Data1 = 0
		local Data2 = 0xFC 
		local Data3 = 0
		local Data4 = 0
		local Status = luup.variable_get(HVAC_USER_SERVICE, "ModeStatus", lul_device)

		if (Status == "Off") then 
			Data1 = 0x08
		else Data1 = 0x18
		end

		local Temp = tonumber(lul_settings['NewCurrentSetpoint'])

		luup.log(BECAENERGY_LOG.."Device " .. lul_device .. ": SetCurrentSetpoint1 " .. Temp)

		if (Temp < lo_temp_limit) then

			Temp = lo_temp_limit

			luup.log(BECAENERGY_LOG.."Temperature setpoint too low")

		elseif (Temp > hi_temp_limit) then

			Temp = hi_temp_limit

			luup.log(BECAENERGY_LOG.."Temperature setpoint too high")

		end

		Temp = tonumber(Temp * 2)

		Data3 = (tonumber(Temp))

		message_type = becaenergy_cmd

		frameSend(frameGenerate(FUNCTION_WRITE, Data1, Data2, Data3, Data4))

		luup.log(BECAENERGY_LOG.."Device " .. lul_device .. ": SetCurrentSetpoint " .. lul_settings['NewCurrentSetpoint'])

	end


	-- Send new temperature hold time (minutes) to BecaEnergy

	function setHoldTime(lul_device,lul_settings)

		local Data1 = 0
		local Data2 = 0xFC 
		local Data3 = 0
		local Data4 = 0
		local Status = luup.variable_get(HVAC_USER_SERVICE, "ModeStatus", lul_device)

		if Status == "Off" then Data1 = 0x08
		else Data1 = 0x18
		end


		local Time = tonumber(lul_settings['HoldTime'])

		if(Time < 0) then

			Time = 0

		elseif(Time > 32000) then

			Time = 0

		end

		Data3 = (bit.band(Time,255))..(bit.band(bit.rshift(Time,8),255))

		message_type = becaenergy_cmd

		frameSend(frameGenerate(FUNCTION_WRITE, Data1, Data2, Data3, Data4))

		luup.log(BECAENERGY_LOG.."Device " .. lul_device .. ": SetHoldTime " .. lul_settings['HoldTime'])

	end


	-- Send new frost temperature setting to BecaEnergy

	function setFrostTemp(lul_device,lul_settings)

		local Data1 = 0
		local Data2 = 0xFC 
		local Data3 = 0
		local Data4 = 0
		local Status = luup.variable_get(HVAC_USER_SERVICE, "ModeStatus", lul_device)

		if Status == "Off" then Data1 = 0x08
		else Data1 = 0x18
		end

		local Temp = tonumber(lul_settings['FrostTemp'])

		if (Temp < lo_frost_limit) then

		Temp = lo_frost_limit

		luup.log(BECAENERGY_LOG.."Frost setpoint too low")

		elseif (Temp > hi_frost_limit) then

			Temp = hi_frost_limit

			luup.log(BECAENERGY_LOG.."Frost setpoint too high")

		end

		Data3 = (Temp * 2)

		message_type = becaenergy_cmd

		frameSend(frameGenerate(FUNCTION_WRITE, Data1, Data2, Data3, Data4))

		luup.log(BECAENERGY_LOG.."Device " .. lul_device .. ": SetFrostTemp " .. lul_settings['FrostTemp'])

	end


	-- Enable Keylock on BecaEnergy

	function setKeyLock(lul_device,lul_settings)

		local Data1 = 0
		local Data2 = 0xFC 
		local Data3 = 0
		local Data4 = 0
		local Status = luup.variable_get(HVAC_USER_SERVICE, "ModeStatus", lul_device)

		if Status == "Off" then Data1 = 0x08
		else Data1 = 0x18
		end

		local State = tonumber(lul_settings['State'])

		if(State ~= 0) then

			State = 1

		end

		Data3 = (State)

		message_type = becaenergy_cmd

		frameSend(frameGenerate(FUNCTION_WRITE, Data1, Data2, Data3, Data4))

		luup.log(BECAENERGY_LOG.."Device " .. lul_device .. ": SetKeyLock " .. lul_settings['State'])

	end


	-- Set BecaEnergy Run Mode (00 = Normal, 01 = Frost Protection)

	function setRunMode(lul_device,lul_settings)

		local Data1 = 0
		local Data2 = 0xFC 
		local Data3 = 0
		local Data4 = 0
		local Status = luup.variable_get(HVAC_USER_SERVICE, "ModeStatus", lul_device)

		if Status == "Off" then Data1 = 0x08
		else Data1 = 0x18
		end

		State = tonumber(lul_settings['State'])

		if(State ~= 0) then

			State = 1

		end

		Data3 = (State)

		message_type = becaenergy_cmd

		frameSend(frameGenerate(FUNCTION_WRITE, Data1, Data2, Data3, Data4))

		luup.log(BECAENERGY_LOG.."Device " .. lul_device .. ": SetRunMode " .. lul_settings['State'])

	end


	-- Set BecaEnergy HW Mode (0 = Auto, 1 = On, 2 = Off)	 ** Note:	PRTHW model only! **

	function setHWMode(lul_device,lul_settings)

		local Data1 = 0
		local Data2 = 0xFC 
		local Data3 = 0
		local Data4 = 0
		local Status = luup.variable_get(HVAC_USER_SERVICE, "ModeStatus", lul_device)

		if Status == "Off" then Data1 = 0x08
		else Data1 = 0x18
		end

		if model ~= 4 then

			luup.log(BECAENERGY_LOG.."Device " .. lul_device .. ": SetHWMode on non-HW model is invalid")

			return false

		end


		local hw_mode = 0

		local new_mode = lul_settings['NewHWMode']

		if new_mode == "On" then hw_mode = 1 end

		if new_mode == "Off" then hw_mode = 2 end

		Data3 = (hw_mode)

		message_type = becaenergy_cmd

		frameSend(frameGenerate(FUNCTION_WRITE, Data1, Data2, Data3, Data4))

		luup.log(BECAENERGY_LOG.."Device " .. lul_device .. ": SetHWMode " .. lul_settings['NewHWMode'])

	end



	-- Set BecaEnergy according to short-cut buttons on device

	function setCurrentMode(lul_device,lul_settings)


		local Data1 = 0
		local Data2 = 0xFC 
		local Data3 = (luup.variable_get(BECAENERGY_SERVICE, "SetHoldTemp", lul_device) * 2)
		local Data4 = 0

		local mode_target = lul_settings['NewModeTarget']

		message_type = becaenergy_cmd
		

		if (mode_target == "Off") then

			-- Set On/Off to 0
			Data1 = 0x08

			frameSend(frameGenerate(FUNCTION_WRITE, Data1, Data2, Data3, Data4))

		
		elseif	(mode_target == "AutoChangeOver") then

			-- Set On/Off to 1

			local Time = 0
			Data1 = 0x18

			frameSend(frameGenerate(FUNCTION_WRITE, Data1, Data2, Data3, Data4))

		
		elseif	(mode_target == "HeatOn") then

			-- Set On/Off to 1

			local Time = 0
			Data1 = 0x18

			local SetTemp = luup.variable_get(BECAENERGY_SERVICE, "SetHoldTemp", lul_device)
			local ActualTemp = luup.variable_get(BECAENERGY_SERVICE, "CurrentTemperature", lul_device)

			-- Set the Hold Temperature to Two degrees above Ambient
			SetTemp = ActualTemp + 2

			if (SetTemp < lo_temp_limit) then

				SetTemp = lo_temp_limit

				luup.log(BECAENERGY_LOG.."Temperature setpoint too low")

			elseif (SetTemp > hi_temp_limit) then

				SetTemp = hi_temp_limit

				luup.log(BECAENERGY_LOG.."Temperature setpoint too high")

			end

			SetTemp = tonumber(SetTemp * 2)

			Data3 = (tonumber(SetTemp))

			message_type = becaenergy_cmd

			frameSend(frameGenerate(FUNCTION_WRITE, Data1, Data2, Data3, Data4))

			luup.log(BECAENERGY_LOG.."Device " .. lul_device .. ": SetCurrentSetpoint " .. lul_settings['NewCurrentSetpoint'])

		end

		luup.variable_set(HVAC_USER_SERVICE, "ModeStatus", mode_target, lul_device)

		luup.log(BECAENERGY_LOG.."Device " .. lul_device .. ": SetCurrentMode " .. mode_target)

	end


	-- Build message to send to BecaEnergy device

	-- Note:	Only send one block of items at a time for now

	function frameGenerate(Function, Data1, Data2, Data3, Data4)

		local ID1 = 0x01
		local ID2 = 0x01
		local CRC = 0

		local OutFrame = ""

		OutFrame = OutFrame..Dec2Hex(Function)
		OutFrame = OutFrame..Dec2Hex(ID1)
		OutFrame = OutFrame..Dec2Hex(ID2)

		OutFrame = OutFrame..Dec2Hex(Data1)
		OutFrame = OutFrame..Dec2Hex(Data2)
		OutFrame = OutFrame..Dec2Hex(Data3)
		OutFrame = OutFrame..Dec2Hex(Data4)

		-- Calculate and then Add the CRC
		CRC = tonumber(Function) + tonumber(ID1) + tonumber(ID2) + tonumber(Data1) + tonumber(Data2) + tonumber(Data3) + tonumber(Data4)
		CRC = bitoper(CRC, 0xFF, AND)
		CRC = bitoper(CRC, 0xA5, XOR)

		OutFrame = OutFrame..Dec2Hex(CRC)

		return OutFrame
	end


	-- Send command to BecaEnergy and receive/process the reply

	function frameSend(OutFrame)

		binCnt= 0

		binData = ""

		for cnt = 1, #OutFrame, 2 do

			val = tonumber(string.sub(OutFrame, cnt, cnt+1),16)

			if 0 == val then
				binData = binData.."\00"

			else
		 		binData = binData..string.format("%c", val)
			end
		end

		becaenergyLogDCB(DCBtoString(binData), "-->", message_type)

		if( luup.io.is_connected(lul_device) == true ) then

			if (luup.io.write(binData) == false ) then
				luup.log(BECAENERGY_LOG.."Can't write to BecaEnergy")
			else
				luup.log(BECAENERGY_LOG.."Data written ok to BecaEnergy")
			end
		
		else

			luup.log(BECAENERGY_LOG.."Can't connect to BecaEnergy")

			comm_failure = comm_failure + 1

			luup.variable_set(HOME_AUTO_SERVICE, "CommFailure", comm_failure, lug_device)

			luup.variable_set(BECAENERGY_SERVICE, "HeatStateText", "COMM FAIL", lug_device)

			-- Set flag to force database update when comms recovers

			cs_change = 1

		end

			end




	-- Read a byte from the DCB

	function dcbReadByte(incomingDCB, start)

		return tonumber (string.byte(string.sub(incomingDCB, start, start)))

	end







	-- Read temperature from DCB and divide by ten to get deg C

	function dcbReadTemp(incomingDCB, start)

		local temp_lo_byte = tonumber (string.byte(string.sub(incomingDCB, start, start)))

		local temp_hi_byte = tonumber (string.byte(string.sub(incomingDCB, start+1, start+1)))

		return (temp_lo_byte + (temp_hi_byte * 256)) / 10

	end







	-- Process received DCB

	function processDCB(incomingDCB)


		local set_temp = 0
		local onoff_status = 0
		local cur_temp = 0
		local hs_change = 0
		local heat_state = 0
		local mode_status = "Off"


		--
		-- Process On/Off Status
		-- 
		onoff_status = bitoper(string.byte(string.sub(incomingDCB, 4, 4)), 0x10, AND)
		if (onoff_status ~= 0) then onoff_status = 1
		end

		if onoff_status ~= last_onoff_status or cs_change ~= 0 then
			luup.variable_set(BECAENERGY_SERVICE, "OnOff", onoff_status, lug_device)
			last_onoff_status = onoff_status
		end

		--
		-- Process HEAT Status
		-- 
		-- Set heating state to display on device dashboard

		if hs_change ~= 0 or cs_change ~= 0 then
			heat_state = bitoper(string.byte(string.sub(incomingDCB, 4, 4)), 0x80, AND)
			if (heat_state ~= 0) then heat_state = 1
			end

			if heat_state ~= last_heat_state or cs_change ~= 0 then
				luup.variable_set(BECAENERGY_SERVICE, "HeatStateText", "HEATING", lug_device)
				last_heat_state = heat_state
			else
				luup.variable_set(BECAENERGY_SERVICE, "HeatStateText", " ", lug_device)
			end
			hs_change = 0
		end			

		--
		-- Process Current Ambient Temperature
		--
		cur_temp = string.byte(string.sub(incomingDCB, 7, 7))
		if ((cur_temp ~=0 and cur_temp ~= last_cur_temp) or cs_change ~= 0) then

			luup.variable_set(TEMPERATURE_SERVICE, "CurrentTemperature", cur_temp/2, lug_device)
			last_cur_temp = cur_temp
		end

		--
		-- Process Set Temperature
		--
		set_temp = string.byte(string.sub(incomingDCB, 6, 6))
		if ((set_temp ~=0 and set_temp ~= last_set_temp) or cs_change ~= 0) then

			luup.variable_set(TEMPERATURE_SETPOINT, "CurrentSetpoint", set_temp/2, lug_device)
			last_set_temp = set_temp
		end

		luup.variable_set(HOME_AUTO_SERVICE, "LastUpdate", os.date("%H:%M:%S"), lug_device)

		
		-- Reset 'force database update' flag after first good data following comms recovery

		cs_change = 0

		--
		-- Set device buttons to reflect actual conditions
		--
		if (onoff_status == 0) then

			mode_status = "Off"

		elseif (onoff_status == 1) and (heat_state == 0) then	

			mode_status = "AutoChangeOver"
		
		elseif (heat_state == 1) then
		
			mode_status = "HeatOn"

		end

		if mode_status ~= last_mode_status then

			luup.variable_set(HVAC_USER_SERVICE, "ModeStatus", mode_status, lug_device)

			last_mode_status = mode_status

		end


	end


		-- Log DCBs to a file

		function becaenergyLogDCB(DCB, direction, msg_type)

			if (loggingEnable == "1") then

			local logfile = '/var/log/cmh/becaenergyDCB.log'

			-- empty file if it reaches 100kb

			local outf = io.open(logfile , 'a')

			local filesize = outf:seek("end")

			outf:close()

			if (filesize > 100000) then

	 				local outf = io.open(logfile , 'w')

	 				outf:write('')

	 				outf:close()

			end

			local outf = io.open(logfile, 'a')

			local mtype = " (" .. message_type .. ") "

			local sep = "	 "

			outf:write(os.date('%d/%m %H:%M:%S'))

			outf:write(sep)

			outf:write(direction)

			outf:write(mtype)

			if DCB == nil then DCB = "?" end

			outf:write(DCB)

			outf:write('\n')

			if direction == "<--" then outf:write('\n') end

			outf:close()

		end

	end

	

	

		-- Convert DCB bytes to a string for logging 

		function DCBtoString(DCB)

		local DCBstr = string.char()

		for i = 1, string.len(DCB) do

			DCBstr = DCBstr .. string.format("%02X ", string.byte(string.sub(DCB, i, i)))

		end

	 	return DCBstr

		end







	-- Convert a value from decimal to hex

	function Dec2Hex(nValue)

		nValue = tonumber(nValue)

		if(nValue == nil) then

			luup.log(BECAENERGY_LOG.."nil in Dec2Hex")

			return "00"

		end

		if(nValue < 16) then

			nHexVal = "0"

		else

		nHexVal = ""

		end

			return nHexVal..string.format("%X", nValue);	-- %X returns uppercase hex, %x gives lowercase letters

	end


	-- Bitwise Operator function
	-- ------------------------
	function bitoper(a, b, oper)
		 local r, m, s = 0, 2^52
		 repeat
			s,a,b = a+b+m, a%m, b%m
			r,m = r + m*oper%(s-a-b), m/2
		 until m < 1
		 return r
	end


	-- BecaEnergy CRC functions
	-- ------------------------
	function CalculateCRC(MyData)

		--becaenergyLogDCB(MyData, "CR1", message_type)

		local MyCRC = (
			string.byte(string.sub(MyData, 1, 1)) + 
			string.byte(string.sub(MyData, 2, 2)) +
			string.byte(string.sub(MyData, 3, 3)) +
			string.byte(string.sub(MyData, 4, 4)) +
			string.byte(string.sub(MyData, 5, 5)) +
			string.byte(string.sub(MyData, 6, 6)) +
			string.byte(string.sub(MyData, 7, 7)) )

		MyCRC = bitoper(MyCRC, 0xFF, AND)
		MyCRC = bitoper(MyCRC, 0xA5, XOR)
					
		return MyCRC

	end







